using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using XNATutorial.Camera.Interfaces;

namespace XNATutorial.Camera
{
    /// <summary>
    /// Modified version from: http://hack1.codeplex.com/releases/view/52280#DownloadId=150242
    /// This is a camera object used to keep track of the position and rotation of a camera. Use of the camera
    /// will affect how models are shown in a scene.
    /// </summary>
    public class QuaternionCamera : ICamera
    {
        #region Properties

        public Vector3 Position { get; set; }
        public Quaternion Rotation { get; protected set; }
        public Matrix Projection { get; protected set; }
        public Matrix View 
        {
            get
            {
                return Matrix.Invert(Matrix.CreateFromQuaternion(Rotation) * Matrix.CreateTranslation(Position));
            }
        }

        #endregion

        #region IMoveableEntity Interface

        /// <summary>
        /// Pans the camera along the X, Y, and Z directions
        /// </summary>
        /// <param name="vecDistance">The amount of X, Y, and Z distance to pan the camera. Units are assumed to be world units.</param>
        public void Translate(Vector3 axis, float fDistance)
        {
            var distance = axis * fDistance;
            Position += Vector3.Transform(distance, Matrix.CreateFromQuaternion(Rotation));
        }

        /// <summary>
        /// Rotates the camera along an arbitrary axis, relative to the position of the camera, not the world.
        /// </summary>
        /// <param name="axis">A Vector3 containing the axis along which to rotate the camera.</param>
        /// <param name="fDegrees">The number of degrees to rotate the camera by.</param>
        public void Rotate(Vector3 axis, float fDegrees)
        {
            axis = Vector3.Transform(axis, Matrix.CreateFromQuaternion(Rotation));

            //Because we use an inverted matrix, we flip the sign on the degrees or the camera will go in
            //the opposite direction that we expect.
            Rotation = Quaternion.Normalize(Quaternion.CreateFromAxisAngle(axis, MathHelper.ToRadians(-fDegrees)) * Rotation);
        }

        #endregion

        #region Constructor & Initialization

        /// <summary>
        /// Constructor of a camera object, placed at the World-coordinate origin and with no rotation.
        /// </summary>
        /// <param name="newViewport">A Viewport object representing the view of the camera.</param>
        public QuaternionCamera(float fieldOfView, float aspectRatio, float nearClippingPlane, float farClippingPlane)
        {
            Position = new Vector3(0, 0, 100);
            Rotation = new Quaternion(0, 0, 0, 1);
            Perspective(fieldOfView, aspectRatio, nearClippingPlane, farClippingPlane);
        }

        /// <summary>
        /// Initializes the projection matrix.
        /// </summary>
        /// <param name="fieldOfView">Field of view.</param>
        /// <param name="aspectRatio">Aspect ratio of the window. See: GraphicsDevice.Viewport</param>
        /// <param name="nearClippingPlane">Near clipping plane of the view frustum.</param>
        /// <param name="farClippingPlane">Far clipping plane of the view frustum.</param>
        public void Perspective(float fieldOfView, float aspectRatio, float nearClippingPlane, float farClippingPlane)
        {
            Projection = Matrix.CreatePerspectiveFieldOfView(fieldOfView, aspectRatio, nearClippingPlane, farClippingPlane);
        }

        #endregion
    }
}
