﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Diagnostics;

namespace XNATutorial.Entity
{
    class ModelEntity : IEntity
    {
        public Model Model { get; set; }

        public Vector3 Position { get; protected set; }
        public Quaternion Rotation { get; protected set; }

        private IEffectMatrices effect;
        public IEffectMatrices Effect
        {
            get
            {
                return effect;
            }
            set
            {
                Debug.Assert(Model != null, "Model Property is not initizalized.");
                
                effect = value;

                // Assign effect to model
                foreach (ModelMesh mesh in Model.Meshes)
                {
                    foreach (ModelMeshPart part in mesh.MeshParts)
                    {
                        Debug.Assert(value is Effect, "Value must be inherited from Effect class.");
                        part.Effect = value as Effect;
                    }
                }
            }
        }

        public ModelEntity(Effect effect, Model model)
        {
            Debug.Assert(effect is IEffectMatrices, "Effect required interface not implemented.");

            Position = Vector3.Zero;
            Rotation = Quaternion.Identity;

            Model = model;
            Effect = effect as IEffectMatrices;
        }

        private Matrix World
        {
            get
            {
                // Howto extract forward- and up-vector from a quaternion:
                // http://www.gamedev.net/community/forums/topic.asp?topic_id=348394
                Vector3 forward = new Vector3(
                    -(2 * Rotation.X * Rotation.Z + 2 * Rotation.Y * Rotation.W),
                    -(2 * Rotation.Y * Rotation.Z - 2 * Rotation.X * Rotation.W),
                    -(1 - 2 * Rotation.X * Rotation.X - 2 * Rotation.Y * Rotation.Y)
                );
                Vector3 up = new Vector3(
                    2 * Rotation.X * Rotation.Y - 2 * Rotation.Z * Rotation.W,
                    1 - 2 * Rotation.X * Rotation.X - 2 * Rotation.Z * Rotation.Z,
                    2 * Rotation.Y * Rotation.Z + 2 * Rotation.X * Rotation.W
                );
                return Matrix.CreateWorld(Position, forward, up);
            }
        }

        public void Draw(GameTime gameTime, Matrix view, Matrix projection)
        {
            Effect.World = World;
            Effect.View = view;
            Effect.Projection = projection;

            foreach (var mesh in Model.Meshes)
            {
                mesh.Draw();
            }
        }

        #region IMoveableEntity Interface

        /// <summary>
        /// Pans the camera along the X, Y, and Z directions
        /// </summary>
        /// <param name="vecDistance">The amount of X, Y, and Z distance to pan the camera. Units are assumed to be world units.</param>
        public void Translate(Vector3 axis, float fDistance)
        {
            Position += axis * fDistance;
        }

        /// <summary>
        /// Rotates the camera along an arbitrary axis, relative to the position of the camera, not the world.
        /// </summary>
        /// <param name="axis">A Vector3 containing the axis along which to rotate the camera.</param>
        /// <param name="fDegrees">The number of degrees to rotate the camera by.</param>
        public void Rotate(Vector3 axis, float fDegrees)
        {
            axis = Vector3.Transform(axis, Matrix.CreateFromQuaternion(Rotation));
            Rotation = Quaternion.Normalize(Quaternion.CreateFromAxisAngle(axis, MathHelper.ToRadians(fDegrees)) * Rotation);
        }

        /// <summary>
        /// Revolves the camera along an axis using a target as the reference point
        /// </summary>
        /// <param name="vecTarget">The target to revolve around</param>
        /// <param name="vecAxis">The axis of rotation</param>
        /// <param name="fDegrees">The degree of rotation</param>
        public void Revolve(Vector3 vecTarget, Vector3 vecAxis, float fDegrees)
        {
        }

        #endregion
    }
}
